# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import logging
import requests
import requests.adapters

from datahub.ingestion.graph.client import DataHubGraph 
from json.decoder import JSONDecodeError
from requests.exceptions import HTTPError, RequestException
from typing import Dict, List, Optional, Tuple, Union

# Class used to fetch Secrets from DataHub. 
class DataHubSecretsClient:

    _graph: DataHubGraph

    def __init__(
        self,
        graph
    ):
        self._graph = graph

    def get_secret_values(self, secret_names: List[str]) -> List[dict[str, str]]:
        if len(secret_names) == 0:
            return {}

        request_json = {
            "query": """query getSecretValues($input: GetSecretValuesInput!) {\n
                getSecretValues(input: $input) {\n
                    name\n
                    value\n
                }\n
            }""",
            "variables": {
                "input": {
                    "secrets": secret_names
                }
            }
        }

        # Fetch secrets using GraphQL API f
        response = self._graph._session.post(
            f"{self._graph.config.server}/api/graphql", json=request_json
        )
        response.raise_for_status()

        # Verify response 
        res_data = response.json()
        if "errors" in res_data:
            raise Exception("Failed to retrieve secrets from DataHub.")

        # Convert list of name, value secret pairs into a dict and return 
        secret_value_list = res_data["data"]["getSecretValues"]
        secret_value_dict = dict()
        for secret_value in secret_value_list:
            secret_value_dict[secret_value["name"]] = secret_value["value"]
        return secret_value_dict


# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pprint

from enum import Enum
from acryl.executor.report.execution_report import ExecutionReport
from acryl.executor.context.execution_context import ExecutionContext
from typing import List, Optional
 
class Type(Enum):
    SUCCESS = 1
    FAILURE = 2
    TIMEOUT = 3
    CANCELLED = 4
    RUNNING = 5

class ExecutionResult:

    # Result type, success or failure
    type: Type

    # Execution context 
    context: ExecutionContext

    def __init__(self, ctx: ExecutionContext) -> None:
        self.context = ctx

    def get_context(self) -> ExecutionContext:
        return self.context
        
    def set_result_type(self, type: Type) -> None:
        self.type = type

    def get_result_type(self) -> Type:
        return self.type 

    def get_summary(self) -> str:
        summary = "~~~~ Execution Summary ~~~~\n\n"
        summary += f"{self.get_context().get_task_name()} - {self.get_context().get_report().as_string()}\n"

        if self.type == Type.FAILURE:
            summary += "Execution finished with errors.\n"
        elif self.type == Type.CANCELLED:
            summary += "Execution was cancelled by request.\n"
        else:
            summary += "Execution finished successfully!\n"
        return summary

    def get_structured_report(self) -> Optional[str]:
        return self.get_context().get_report().structured_report


    def pretty_print_summary(self):
        print(self.get_summary())


# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import Dict, List, Union
from acryl.executor.common.config import ConfigModel

class SecretStoreConfig(ConfigModel):
    type: str
    config: Dict 

# Abstract base class for a Secret Store, or a class that resolves "secret" values 
# by name. 
class SecretStore:

    @classmethod
    @abstractmethod
    def create(cls, configs: dict) -> "SecretStore":
        pass

    @abstractmethod
    def get_secret_values(self, secret_names: List[str]) -> Dict[str, Union[str, None]]:
        """
        Attempt to fetch a group of secrets, returning a Dictionary of the secret of None if one 
        cannot be resolved by the store. 
        """

    @abstractmethod
    def get_secret_value(self, secret_name: str) -> Union[str, None]:
        """
        Attempt to fetch a secret, or return None if one cannot be resolved. 
        """

    @abstractmethod
    def get_id(self) -> str:
        """
        Get a unique name or id associated with the Secret Store. 
        """

    @abstractmethod
    def close(self) -> None:
        """
        Wraps up the task
        """


# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
import signal

from acryl.executor.execution.default_executor import DefaultExecutor, DefaultExecutorConfig
from acryl.executor.execution.task import TaskConfig
from acryl.executor.request.signal_request import SignalRequest
from acryl.executor.request.execution_request import ExecutionRequest
from acryl.executor.result.execution_result import ExecutionResult, Type
from acryl.executor.secret.secret_store import SecretStoreConfig
from datahub.metadata.schema_classes import ExecutionRequestResultClass, ExecutionRequestKeyClass, ExecutionRequestSignalClass, StructuredExecutionReportClass
from datahub.metadata.com.linkedin.pegasus2avro.mxe import (
    GenericAspectClass
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.serialization_helper import pre_json_transform
from datahub.ingestion.graph.client import DataHubGraph, DatahubClientConfig
from typing import List, Optional
from pydantic import validator

DATAHUB_EXECUTION_REQUEST_ENTITY_NAME = "dataHubExecutionRequest"
DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME = "dataHubExecutionRequestResult"

class ReportingExecutorConfig(DefaultExecutorConfig):
    id: str = "default"
    task_configs: List[TaskConfig]
    secret_stores: List[SecretStoreConfig]

    graph_client: Optional[DataHubGraph] = None
    graph_client_config: Optional[DatahubClientConfig] = None

    class Config:
        arbitrary_types_allowed = True

    @validator('graph_client')
    def check_graph_connection(cls, v):
        if v is not None:
            v.test_connection()
        return v 

# Supports RUN_INGEST commands. 
class ReportingExecutor(DefaultExecutor):

    _datahub_graph: DataHubGraph
    _execution_timeout_sec: Optional[float] = 86400 # Default timeout is 1 day 

    def __init__(self, exec_config: ReportingExecutorConfig) -> None:
        super().__init__(exec_config)
        if exec_config.graph_client is not None:
            self._datahub_graph = exec_config.graph_client
        elif exec_config.graph_client_config is not None:
            self._datahub_graph = DataHubGraph(exec_config.graph_client_config)
        else:
            raise Exception("Invalid configuration provided. Missing DataHub graph client configs")
    
    # Run a list of tasks in sequence 
    def execute(self, request: ExecutionRequest) -> ExecutionResult:

        # Capture execution start time 
        start_time_ms = round(time.time() * 1000)

        # Build & emit an ACK mcp
        kickoff_mcp = self._build_kickoff_mcp(request, start_time_ms)
        self._datahub_graph.emit_mcp(kickoff_mcp)

        # Execute the request  
        request.progress_callback = lambda partial_report: self._datahub_graph.emit_mcp(self._build_progress_mcp(request, start_time_ms, partial_report=partial_report))
        exec_result = super().execute(request)

        # Execution has completed. Report the status. Only consider the first task in the list. 
        # TODO: Support timed out state. 
        status = exec_result.type.name

        # Capture the report
        report = exec_result.get_summary()

        # Get a structured report if available
        structured_report = exec_result.get_structured_report()

        # Compute execution duration (wall clock)
        end_time_ms = round(time.time() * 1000)
        duration_ms = end_time_ms - start_time_ms

        # Build & emit the completion mcp 
        completion_mcp = self._build_completion_mcp(exec_request=request, status=status, start_time_ms=start_time_ms, duration_ms=duration_ms, report=report, structured_report=structured_report)
        self._datahub_graph.emit_mcp(completion_mcp)

        return exec_result

    def signal(self, request: SignalRequest) -> None: 
        super().signal(request)
        # If the task cannot be found (is not actively executing), emit an empty cancellation event. 
        if request.signal == 'KILL':
            task_futures = self.task_futures.get(request.exec_id)
            if task_futures is None:
                # No task found. Simply emit a cancelled MCE. The start time is unclear. 
                # Build & emit the cancellation mcp 
                cancellation_mcp = self._build_empty_cancel_mcp(exec_id=request.exec_id)
                self._datahub_graph.emit_mcp(cancellation_mcp)
            
    # Builds an MCP to report the start of execution request handling 
    def _build_kickoff_mcp(self, exec_request: ExecutionRequest, start_time_ms: int) -> MetadataChangeProposalWrapper: 
        assert exec_request.exec_id, f"Missing exec_id for request {exec_request}"
        key_aspect = self._build_execution_request_key_aspect(exec_request.exec_id)
        result_aspect = self._build_execution_request_result_aspect(
            status=Type.RUNNING.name,
            start_time_ms=start_time_ms
        )

        return MetadataChangeProposalWrapper(
            entityType=DATAHUB_EXECUTION_REQUEST_ENTITY_NAME,
            changeType="UPSERT",
            auditHeader=None,
            entityKeyAspect=key_aspect,
            aspectName=DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME,
            aspect=result_aspect
        )

    def _build_progress_mcp(self, exec_request: ExecutionRequest, start_time_ms: int, partial_report: str) -> MetadataChangeProposalWrapper:
        assert exec_request.exec_id, f"Missing exec_id for request {exec_request}"
        key_aspect = self._build_execution_request_key_aspect(exec_request.exec_id)
        result_aspect = self._build_execution_request_result_aspect(
            status=Type.RUNNING.name,
            start_time_ms=start_time_ms,
            report=partial_report,
        )

        return MetadataChangeProposalWrapper(
            entityType=DATAHUB_EXECUTION_REQUEST_ENTITY_NAME,
            changeType="UPSERT",
            auditHeader=None,
            entityKeyAspect=key_aspect,
            aspectName=DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME,
            aspect=result_aspect
        )

    # Builds an MCP to report the completion of execution request handling 
    def _build_completion_mcp(self, exec_request: ExecutionRequest, status: str, start_time_ms: int, duration_ms: int, report: str, structured_report: Optional[str]) -> MetadataChangeProposalWrapper: 
        key_aspect = self._build_execution_request_key_aspect(exec_request.exec_id or "missing execution id")
        result_aspect = self._build_execution_request_result_aspect(
            status=status,
            start_time_ms=start_time_ms,
            duration_ms=duration_ms,
            report=report,
            structured_report=StructuredExecutionReportClass(type=exec_request.name, serializedValue=structured_report, contentType="application/json") if structured_report else None
        )

        return MetadataChangeProposalWrapper(
            entityType=DATAHUB_EXECUTION_REQUEST_ENTITY_NAME,
            changeType="UPSERT",
            auditHeader=None,
            entityKeyAspect=key_aspect,
            aspectName=DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME,
            aspect=result_aspect
        )

    # Builds an MCP to report the completion of execution request handling 
    def _build_empty_cancel_mcp(self, exec_id: str) -> MetadataChangeProposalWrapper: 
        key_aspect = self._build_execution_request_key_aspect(exec_id)

        # TODO: Determine whether this is the "right" thing to do.
        result_aspect = self._build_execution_request_result_aspect(
            status=Type.CANCELLED.name,
            start_time_ms=0, # TODO: Make start time optional
            duration_ms=None,
            report="No active execution request found."
        )

        return MetadataChangeProposalWrapper(
            entityType=DATAHUB_EXECUTION_REQUEST_ENTITY_NAME,
            changeType="UPSERT",
            auditHeader=None,
            entityKeyAspect=key_aspect,
            aspectName=DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME,
            aspect=result_aspect
        )

    def _build_execution_request_key_aspect(self, execution_request_id: str) -> ExecutionRequestKeyClass:
        # Build the JSON map
        execution_request_key: ExecutionRequestKeyClass = ExecutionRequestKeyClass(id=execution_request_id)
        return execution_request_key

    def _build_execution_request_result_aspect(
        self, 
        status: str, 
        start_time_ms: int, 
        duration_ms: Optional[int] = None,
        report: Optional[str] = None,
        structured_report: Optional[StructuredExecutionReportClass] = None
        ) -> ExecutionRequestResultClass:

        execution_request_result: ExecutionRequestResultClass = ExecutionRequestResultClass(
            status=status,
            startTimeMs=start_time_ms,
            durationMs=duration_ms,
            report=report,
            structuredReport=structured_report
        )
        return execution_request_result




# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from acryl.executor.secret.secret_store import SecretStore
from acryl.executor.secret.datahub_secrets_client import DataHubSecretsClient
from datahub.ingestion.graph.client import DataHubGraph, DatahubClientConfig
from pydantic import BaseModel, validator
from typing import Union, Dict, List, Optional

logger = logging.getLogger(__name__)

class DataHubSecretStoreConfig(BaseModel):

    graph_client: Optional[DataHubGraph] = None
    graph_client_config: Optional[DatahubClientConfig] = None

    class Config:
        arbitrary_types_allowed = True

    @validator('graph_client')
    def check_graph_connection(cls, v):
        if v is not None:
            v.test_connection()
        return v


# An implementation of SecretStore that fetches secrets from DataHub
class DataHubSecretStore(SecretStore):

    # Client for fetching secrets from DataHub GraphQL API 
    client: DataHubSecretsClient

    def __init__(self, config):
        # Attempt to establish an outbound connection to DataHub and create a client.
        if config.graph_client is not None:
            self.client = DataHubSecretsClient(
                graph=config.graph_client
            )
        elif config.graph_client_config is not None: 
            graph = DataHubGraph(config.graph_client_config)
            self.client = DataHubSecretsClient(graph)
        else:
            raise Exception("Invalid configuration provided: unable to construct DataHub Graph Client.")

    def get_secret_values(self, secret_names: List[str]) -> Dict[str, Union[str, None]]:
        # Fetch the secret from DataHub, using the credentials provided in the configuration. 
        # Use the GraphQL API. 
        try:
            return self.client.get_secret_values(secret_names)
        except Exception:
            # Failed to resolve secrets, return empty. 
            logger.exception(f"Caught exception while attempting to fetch secrets from DataHub. Secret names: {secret_names}")
            return {}

    def get_secret_value(self, secret_name: str) -> Union[str, None]:
        secret_value_dict = self.get_secret_values([secret_name])
        return secret_value_dict.get(secret_name)

    def get_id(self) -> Union[str, None]:
        return "datahub"

    @classmethod
    def create(cls, config):
        config = DataHubSecretStoreConfig.parse_obj(config)
        return cls(config)




# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pprint
import datetime 

from dataclasses import dataclass, field
from typing import List, Dict, Optional

def format_report_line(line_type: str, headers: List[str], message: str) -> str:
    utc_time = datetime.datetime.utcnow()
    header_str = ""
    for header in headers:
        header_str = header_str + f"[{header}]"
    return f"{utc_time} {header_str} {line_type}: {message}"

@dataclass
class Report:
    def as_obj(self) -> dict:
        return self.__dict__

    def as_string(self) -> str:
        return pprint.pformat(self.as_obj(), width=150)

    def as_json(self) -> str:
        return json.dumps(self.as_obj())

"""
Report object leveraged by an Executor coordinating Task Execution.
"""
@dataclass
class ExecutionReport(Report):

    exec_id: str

    infos: List[str] = field(default_factory=list)
    errors: List[str] = field(default_factory=list)
    structured_report: Optional[str] = None

    def __init__(self, exec_id: str): 
        self.exec_id = exec_id
        self.infos = []
        self.errors = []

    def report_info(self, message: str) -> None:
        formatted_message = self._format_report_line("INFO", message)
        print(formatted_message) # TODO: Add flag for printing to STDOUT
        self.infos.append(formatted_message)

    def report_error(self, message: str) -> None:
        formatted_message = self._format_report_line("ERROR", message)
        print(formatted_message) # TODO: Add flag for printing to STDOUT
        self.errors.append(formatted_message)

    def set_structured_report(self, report_content: str) -> None:
        self.structured_report = report_content

    def _format_report_line(self, line_type: str, message: str) -> str:
        headers = [
            f"exec_id={self.exec_id}"
        ]
        return format_report_line(line_type, headers, message)


# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import traceback
import asyncio
import logging
import nest_asyncio

from typing import Any, Dict, List
from acryl.executor.common.config import ConfigModel
from acryl.executor.execution.executor import Executor
from acryl.executor.report.execution_report import ExecutionReport
from acryl.executor.context.executor_context import ExecutorContext
from acryl.executor.context.execution_context import ExecutionContext
from acryl.executor.execution.task import Task, TaskConfig
from acryl.executor.execution.task_registry import TaskRegistry
from acryl.executor.request.execution_request import ExecutionRequest
from acryl.executor.result.execution_result import ExecutionResult, Type
from acryl.executor.request.signal_request import SignalRequest
from acryl.executor.secret.secret_store import SecretStore, SecretStoreConfig
from acryl.executor.secret.secret_store_registry import SecretStoreRegistry

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

"""
Sets up the Task Executor 
"""
class DefaultExecutorConfig(ConfigModel):
    id: str
    task_configs: List[TaskConfig]
    secret_stores: List[SecretStoreConfig] = []

"""
Executes a Task Execution Request 
"""
class DefaultExecutor(Executor):
    id: str 

    # Execution Event Loop
    event_loop = asyncio.new_event_loop()

    # Patch the primary event loop to permit multiple runs 
    nest_asyncio.apply(event_loop)


    # Tasks
    task_registry: TaskRegistry = TaskRegistry()
    task_instances: Dict[str, Task] = {}
    task_futures: Dict[str, List[Any]] = {}

    # Secret Stores 
    secret_stores: List[SecretStore] = []

    def __init__(self, config: DefaultExecutorConfig) -> None:
        # Register the tasks 
        self.id = config.id
        self.secret_stores = self._create_secret_stores(config)
        for task_config in config.task_configs:
            self.register_task(task_config)

        # Set event loop context 
        asyncio.set_event_loop(self.event_loop)

    def register_task(self, task: TaskConfig) -> None:
        # Store a registry of command -> task. 
        self.task_registry.register_lazy(task.name, task.type)

        # Create and initialize the task instances
        task_class = self.task_registry.get(task.name)
        try: 
            executor_context = ExecutorContext(self.id, self.secret_stores)
            task_instance = task_class.create(task.configs, executor_context)
            self.task_instances[task.name] = task_instance
        except Exception:
            raise Exception(f"Failed to create instance of task with name {task.name}: {traceback.format_exc(limit=3)}")

    # Run a list of tasks in sequence 
    def execute(self, request: ExecutionRequest) -> ExecutionResult:

        # 1. Create execution context
        executor_context = ExecutorContext(executor_id=self.id, secret_stores=self.secret_stores)

        # 2. Execute the task in the request 
        return self.execute_task(request, executor_context)

    def execute_task(self, request: ExecutionRequest, executor_context: ExecutorContext) -> ExecutionResult:

        # 1. Build task execution context
        execution_context = ExecutionContext(request)
        execution_id = execution_context.get_execution_id()
        execution_report = execution_context.get_report()
        execution_report.report_info(f"Starting execution for task with name={request.name}")

        # 2. Retrieve an instance of the task  
        task_instance = self.task_instances.get(request.name)
        execution_result = ExecutionResult(execution_context)
        execution_result.set_result_type(Type.FAILURE)
        if task_instance is None:
            execution_report.report_info(f"Failed to find task with name={request.name}")
            return execution_result 

        # 3. Execute task - task is expected to throw TaskError on failure. 
        try:
            # Execute task via an event loop. Store the executing task such that it can be cancelled. 
            # TODO: Have an async version of this method. 
            task_future = self.event_loop.create_task(task_instance.execute(request.args, execution_context))
            logger.debug(f"Task for {execution_id} created")
            self.task_futures[execution_id] = [task_future]
            self.event_loop.run_until_complete(task_future)
            logger.debug(f"Task for {execution_id} completed")
        except Exception:
            execution_report.report_info(f"Caught exception EXECUTING task_id={execution_context.get_execution_id()}, name={request.name}, stacktrace={traceback.format_exc(limit=20)}")
            return execution_result 
        except asyncio.exceptions.CancelledError:
            execution_report.report_info(f"Execution cancelled while EXECUTING task_id={execution_context.get_execution_id()}, name={request.name}, stacktrace={traceback.format_exc(limit=3)}")
            execution_result.set_result_type(Type.CANCELLED)
            return execution_result 

        execution_result.set_result_type(Type.SUCCESS)
        return execution_result

    def signal(self, request: SignalRequest) -> None: 
        if request.signal == 'KILL':
            task_futures = self.task_futures.get(request.exec_id)
            if task_futures is None:
                logger.error(f"Received KILL for missing task ID {request.exec_id}")
            else:
                for task_future in task_futures: 
                    # Cancel the task if not complete. 
                    if not task_future.done():
                        logger.debug(f"Trying to cancel {task_future}")
                        task_future.cancel()
                

    def shutdown(self) -> None:
        for task_name, task_instance in self.task_instances.items():
            try:
                task_instance.close()
            except Exception:
                logger.warn(f"Failed to shutdown task with name {task_name}")

    def get_id(self) -> str:
        return self.id
    
    def _create_secret_stores(self, config: DefaultExecutorConfig) -> List[SecretStore]:
        secret_stores = [] 
        for secret_store_config in config.secret_stores:
            secret_stores.append(self._create_secret_store(secret_store_config))
        return secret_stores

    def _create_secret_store(self, config: SecretStoreConfig) -> SecretStore:
        # Create a secret store registry 
        secret_store_registry: SecretStoreRegistry = SecretStoreRegistry()

        # Get the secret store configs
        secret_store_type: str = config.type
        secret_store_configs: dict = config.config 

        # Fetch the correct secret store, or register a new one
        if not secret_store_registry.is_enabled(secret_store_type):
            # Custom Secret Store found. Register it. 
            secret_store_registry.register_lazy(secret_store_type, secret_store_type)
        
        # Instantiate the secret store class
        secret_store_class = secret_store_registry.get(secret_store_type)

        # Create & return new instance of secret store 
        return secret_store_class.create(config.config)


# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
import traceback

from acryl.executor.dispatcher.dispatcher import Dispatcher
from acryl.executor.execution.executor import Executor
from acryl.executor.request.execution_request import ExecutionRequest 
from acryl.executor.request.signal_request import SignalRequest 
from typing import List

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

def dispatch_async(executor: Executor, request: ExecutionRequest):
    try:
        res = executor.execute(request)
        res.pretty_print_summary()
    except Exception:
        logger.error(f"Failed dispatch for {request.exec_id}: {traceback.format_exc(limit=3)}")

def dispatch_signal_async(executor: Executor, request: SignalRequest):
    try:
        executor.signal(request)
    except Exception:
        logger.error(f"Failed signal dispatch for {request.exec_id}: {traceback.format_exc(limit=3)}")

# An abstract base class representing a dispatcher capable of dispatching Execution Requests 
class DefaultDispatcher(Dispatcher):

    def __init__(self, executors: List[Executor]):
        self.executors: List[Executor] = executors
        self.threads: List[threading.Thread] = []


    def dispatch(self, request: ExecutionRequest) -> None: 
        # Determine which executor should handle the request. 
        for executor in self.executors:
            if executor.get_id() == request.executor_id:
                # Simply execute the task on a new thread.
                thread = threading.Thread(target=dispatch_async, args=(executor, request))
                self.threads.append(thread)
                thread.start()
                logger.debug(f"Started thread {thread} for {request.exec_id}")
                return

        raise Exception(f'Failed to find executor {request.executor_id} for Execution Request with execution id {request.exec_id}')

    def dispatch_signal(self, request: SignalRequest) -> None: 
        # Determine which executor should handle the signal. 
        for executor in self.executors:
            if executor.get_id() == request.executor_id:
                # Simply execute the task on a new thread.
                thread = threading.Thread(target=dispatch_signal_async, args=(executor, request))
                thread.start()
                logger.debug(f"Started signal thread {thread} for {request.exec_id}")
                return

        raise Exception(f'Failed to find executor {request.executor_id} for Signal Request with execution id {request.exec_id}.')

    def shutdown(self):
        for t in self.threads:
            t.join()



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List
import os
import shutil
import yaml
import json
from typing import Sequence
import re
from acryl.executor.context.executor_context import ExecutorContext
from acryl.executor.context.execution_context import ExecutionContext
from acryl.executor.execution.task import TaskError

logger = logging.getLogger(__name__)


class SubProcessTaskUtil:

    MAX_LOG_LINES = 2000

    @staticmethod
    def _format_log_lines(lines: Sequence[str]) -> str: 
        text = "".join(lines)
        if len(lines) > SubProcessTaskUtil.MAX_LOG_LINES:
            text = f"[earlier logs truncated...]\n{text}"

        return text

    @staticmethod
    def _resolve_secrets(secret_names: List[str], ctx: ExecutorContext) -> dict:
        # Attempt to resolve secret using by checking each configured secret store.
        secret_stores = ctx.get_secret_stores()
        final_secret_values = dict({})

        for secret_store in secret_stores:
            try: 
                # Retrieve secret values from the store. 
                secret_values_dict = secret_store.get_secret_values(secret_names)
                # Overlay secret values from each store, if not None. 
                for secret_name, secret_value in secret_values_dict.items():
                    if secret_value is not None:
                        final_secret_values[secret_name] = secret_value
            except Exception:
                logger.exception(f"Failed to fetch secret values from secret store with id {secret_store.get_id()}")
        return final_secret_values

    @staticmethod
    def _resolve_recipe(recipe: str, execution_ctx: ExecutionContext, executor_ctx: ExecutorContext) -> dict:
        # Now attempt to find and replace all secrets inside the recipe. 
        secret_pattern = re.compile('.*?\${(\w+)}.*?')

        resolved_recipe = recipe
        secret_matches = secret_pattern.findall(resolved_recipe)

        # 1. Extract all secrets needing resolved. 
        secrets_to_resolve = []
        if secret_matches:
            for match in secret_matches:
                secrets_to_resolve.append(match)

        # 2. Resolve secret values
        secret_values_dict = SubProcessTaskUtil._resolve_secrets(secrets_to_resolve, executor_ctx)

        # 3. Substitute secrets into recipe file
        if secret_matches:
            for match in secret_matches:
                # a. Check if secret was successfully resolved.
                secret_value = secret_values_dict.get(match)
                if secret_value is None:
                    # Failed to resolve secret. 
                    raise TaskError(f"Failed to resolve secret with name {match}. Aborting recipe execution.")
                    
                # b. Substitute secret value. 
                resolved_recipe = resolved_recipe.replace(
                    f'${{{match}}}', secret_value
                )

        json_recipe = json.loads(resolved_recipe)

        # Inject run_id into the recipe
        json_recipe["run_id"] = execution_ctx.exec_id

        # TODO: Inject the ingestion source id as well.

        # For now expect that the recipe is complete, this may not be the case, however for hybrid deployments. Secret store!
        return json_recipe

    @staticmethod
    def _write_recipe_to_file(dir_path: str, file_name: str, recipe: dict) -> None:

        # 1. Create directories to the path
        os.makedirs(dir_path, mode = 0o777, exist_ok = True)

        # 2. Dump recipe dictionary to a YAML string
        yaml_recipe = yaml.dump(recipe)

        # 3. Write YAML to file  
        file_handle = open(dir_path + "/" + file_name, "w")
        n = file_handle.write(yaml_recipe)
        file_handle.close()

    @staticmethod
    def _remove_directory(dir_path: str) -> None:
        shutil.rmtree(dir_path)


# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import subprocess
import os
import logging
import asyncio
from collections import deque
from acryl.executor.execution.task import Task
from acryl.executor.execution.task import TaskError
from acryl.executor.context.executor_context import ExecutorContext
from acryl.executor.context.execution_context import ExecutionContext
from acryl.executor.common.config import ConfigModel
from acryl.executor.execution.sub_process_task_common import SubProcessTaskUtil
from typing import List

logger = logging.getLogger(__name__)

class SubProcessTestConnectionTaskConfig(ConfigModel):
    tmp_dir: str = "/tmp/datahub/ingest"

class SubProcessTestConnectionTaskArgs(ConfigModel):
    recipe: str
    version: str = "latest"

class SubProcessTestConnectionTask(Task):

    config: SubProcessTestConnectionTaskConfig
    tmp_dir: str # Location where tmp files will be written (recipes) 
    ctx: ExecutorContext

    @classmethod
    def create(cls, config: dict, ctx: ExecutorContext) -> "Task":
        config_parsed = SubProcessTestConnectionTaskConfig.parse_obj(config)
        return cls(config_parsed, ctx)

    def __init__(self, config: SubProcessTestConnectionTaskConfig, ctx: ExecutorContext):
        self.config = config
        self.tmp_dir = config.tmp_dir
        self.ctx = ctx 

    async def execute(self, args: dict, ctx: ExecutionContext) -> None:

        exec_id = ctx.exec_id # The unique execution id.

        exec_out_dir = f"{self.tmp_dir}/{exec_id}"

        # 0. Validate arguments 
        validated_args = SubProcessTestConnectionTaskArgs.parse_obj(args)

        # 1. Resolve the recipe (combine it with others)
        recipe: dict = SubProcessTaskUtil._resolve_recipe(validated_args.recipe, execution_ctx=ctx, executor_ctx=self.ctx)

        # 2. Write recipe file to local FS (requires write permissions to /tmp directory)
        recipe_file: str = "recipe.yml"  
        SubProcessTaskUtil._write_recipe_to_file(exec_out_dir, recipe_file, recipe)
        recipe_file_loc: str = f"{exec_out_dir}/{recipe_file}"

        # 3. Spin off subprocess to run the run_ingest.sh script
        datahub_version = validated_args.version # The version of DataHub CLI to use. 
        plugins = recipe["source"]["type"] # The source type -- ASSUMPTION ALERT: This should always correspond to the plugin name. 
        command_script: str = "run_test_connection.sh" # TODO: Make sure this is EXECUTABLE.  
        report_out_file: str = f"{exec_out_dir}/connection_report.json"
        stdout_lines: deque = deque(maxlen=SubProcessTaskUtil.MAX_LOG_LINES)

        # TODO: Inject a token into the recipe to run such that it "just works" (currently, a token is required)

        ingest_process = subprocess.Popen([command_script, exec_id, datahub_version, plugins, self.tmp_dir, recipe_file_loc, report_out_file], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

        try: 
            while ingest_process.poll() is None:
                assert ingest_process.stdout
                line = ingest_process.stdout.readline()

                sys.stdout.write(line)
                stdout_lines.append(line)
                await asyncio.sleep(0)

            return_code = ingest_process.poll()

        except asyncio.CancelledError:
            # Terminate the running child process 
            ingest_process.terminate()
            raise  

        finally:
            if os.path.exists(report_out_file):
                with open(report_out_file,'r') as structured_report_fp:
                    ctx.get_report().set_structured_report(structured_report_fp.read())

            ctx.get_report().report_info(f"stdout={SubProcessTaskUtil._format_log_lines(stdout_lines)}")
            
            # Cleanup by removing the exec out directory
            SubProcessTaskUtil._remove_directory(exec_out_dir)

        if return_code != 0:
            # Failed
            ctx.get_report().report_info("Failed to execute 'datahub test connection'")
            raise TaskError("Failed to execute 'datahub test connection'") 
        
        # Report Successful execution
        ctx.get_report().report_info("Successfully executed 'datahub test connection'")


    def close(self) -> None:
        pass


# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import json
import os
from unittest import mock


from acryl.executor.request.execution_request import ExecutionRequest
from acryl.executor.request.signal_request import SignalRequest
from acryl.executor.execution.task import TaskConfig
from acryl.executor.dispatcher.default_dispatcher import DefaultDispatcher
from acryl.executor.execution.reporting_executor import ReportingExecutor, ReportingExecutorConfig
from acryl.executor.secret.datahub_secret_store import DataHubSecretStoreConfig
from acryl.executor.secret.secret_store import SecretStoreConfig
from datahub.ingestion.graph.client import DatahubClientConfig, DataHubGraph
from datahub.ingestion.sink.file import FileSink, FileSinkConfig
from datahub.ingestion.api.sink import NoopWriteCallback
# this is an integration test of the functionality, basically, manually verified  
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.common import RecordEnvelope
   

class FakeDataHubGraph(DataHubGraph):
    def __init__(self, file_name: str = "mcp_logs.json"):
        self.file_name = file_name
        self.file_sink = FileSink(PipelineContext(run_id="test"), FileSinkConfig(filename=file_name))

    def test_connection(self) -> dict:
        return {"noCode": True}

    def emit_mcp(self, mcp):
        self.file_sink.write_record_async(RecordEnvelope(record=mcp, metadata={}), write_callback=NoopWriteCallback())

    def close(self):
        self.file_sink.close()

def setup():
    # Main function

    env_map = {"HOST_PORT": "http://localhost:8080",
                "DATABASE": "datahub",
                "MYSQL_USERNAME": "datahub",
                "MYSQL_PASSWORD": "datahub",
            }
    for k,v in env_map.items():
        os.environ[k] = v

    fake_client = FakeDataHubGraph()
    # Build and register executors with the dispatcher
    local_task_config = TaskConfig(
        name="RUN_INGEST",
        type="acryl.executor.execution.sub_process_ingestion_task.SubProcessIngestionTask", 
        configs=dict({}))
    local_test_connection_task_config = TaskConfig(
        name="TEST_CONNECTION",
        type="acryl.executor.execution.sub_process_test_connection_task.SubProcessTestConnectionTask",
        configs=dict({})
    )
    local_exec_config = ReportingExecutorConfig(
        id="default", 
        task_configs=[local_task_config, local_test_connection_task_config],
        secret_stores=[
            SecretStoreConfig(type="env", config=dict({})),
            SecretStoreConfig(type="datahub", config=DataHubSecretStoreConfig(
                graph_client_config=DatahubClientConfig(
                    server="http://localhost:8080",
                    token="eyJhbGciOiJIUzI1NiJ9.eyJhY3RvclR5cGUiOiJVU0VSIiwiYWN0b3JJZCI6ImRhdGFodWIiLCJ0eXBlIjoiUEVSU09OQUwiLCJ2ZXJzaW9uIjoiMSIsImV4cCI6MTY0MjIyMzE5MywianRpIjoiZjIwZjRmMmQtMThlNS00NTRmLTgyMzQtMTg1YzIxZjAwN2QzIiwic3ViIjoiZGF0YWh1YiIsImlzcyI6ImRhdGFodWItbWV0YWRhdGEtc2VydmljZSJ9.e3N4_Z_3QhviYmLW8UXS2KhhcrBIMinDhSjnR05bMbE"
                )
            ))
        ],
        graph_client_config=DatahubClientConfig(
            server="http://localhost:8080",
            token="eyJhbGciOiJIUzI1NiJ9.eyJhY3RvclR5cGUiOiJVU0VSIiwiYWN0b3JJZCI6ImRhdGFodWIiLCJ0eXBlIjoiUEVSU09OQUwiLCJ2ZXJzaW9uIjoiMSIsImV4cCI6MTY0MjIyMzE5MywianRpIjoiZjIwZjRmMmQtMThlNS00NTRmLTgyMzQtMTg1YzIxZjAwN2QzIiwic3ViIjoiZGF0YWh1YiIsImlzcyI6ImRhdGFodWItbWV0YWRhdGEtc2VydmljZSJ9.e3N4_Z_3QhviYmLW8UXS2KhhcrBIMinDhSjnR05bMbE"
        ),
        graph_client=fake_client
    )
    return (local_task_config, local_exec_config, fake_client, local_test_connection_task_config)


if __name__ == "__main__":

    (local_task_config, local_exec_config, fake_client, local_test_connection_task_config) = setup()
    local_executor = ReportingExecutor(local_exec_config)
    dispatcher = DefaultDispatcher([local_executor])

    endpoint = "http://localhost:8080/" # get from config
    recipe_dictionary = {
        "source": {
            "type": "mysql",
            "config": {
                "filename": "foobar.yaml"
                }
            },
        "sink": {
            "type": "datahub-rest",
            "config": {
                "server":"http://localhost:8080", 
                "token":"eyJhbGciOiJIUzI1NiJ9.eyJhY3RvclR5cGUiOiJVU0VSIiwiYWN0b3JJZCI6ImRhdGFodWIiLCJ0eXBlIjoiUEVSU09OQUwiLCJ2ZXJzaW9uIjoiMSIsImV4cCI6MTY0MjIyMzE5MywianRpIjoiZjIwZjRmMmQtMThlNS00NTRmLTgyMzQtMTg1YzIxZjAwN2QzIiwic3ViIjoiZGF0YWh1YiIsImlzcyI6ImRhdGFodWItbWV0YWRhdGEtc2VydmljZSJ9.e3N4_Z_3QhviYmLW8UXS2KhhcrBIMinDhSjnR05bMbE"
                }
        }
    }
    args = {'version': "0.8.41.1rc2",
            'recipe': json.dumps(recipe_dictionary)
    }

    exec_request = ExecutionRequest(exec_id="1234", name="TEST_CONNECTION", args=args)
    start_time = time.time()
    dispatcher.dispatch(exec_request)
    
    dispatcher.shutdown()
    fake_client.close()
    # Now kill the process. 
    # signal_request = SignalRequest(exec_id="1234", signal="KILL")
    # dispatcher.dispatch_signal(signal_request)

    


